package main

import (
	"bufio"
	"bytes"
	"encoding/xml"
	"errors"
	"fmt"
	"io"
	"log"
	"net/http"
	"net/url"
	"strings"
	"text/template"
	"time"
)

const ICAL_TIME = "20060102T150405"
const ICAL_DATE = "20060102"

type CalDAV struct {
	ServerSettings
	httpClient *http.Client
}

type ICal struct {
	data      []string
	eventData []string // sub slice
}

func (i ICal) Summary() string {
	for _, s := range i.eventData {
		if strings.HasPrefix(s, "SUMMARY:") {
			return s[len("SUMMARY:"):]
		}
	}
	return ""
}

func (i ICal) UID() string {
	for _, s := range i.eventData {
		if strings.HasPrefix(s, "UID:") {
			return s[len("UID:"):]
		}
	}
	return ""
}

func (i ICal) getTimeField(f string) time.Time {
	for _, s := range i.eventData {
		if strings.HasPrefix(s, f) {
			st := s[strings.LastIndex(s, ":")+1:]
			// BEWARE, we don't parse possible ICal timezones (yet). If it is not
			// a UTC time, we simply assume local time.
			zone := time.Local
			if st[len(st)-1:] == "Z" {
				st = st[:len(st)-1]
				zone = time.UTC
			}
			t, err := time.ParseInLocation(ICAL_TIME, st, zone)
			if err != nil {
				log.Printf("Cannot decode %s '%s': %s", f, st, err)
			}
			return t
		}
	}
	return time.Time{}
}

func (i ICal) Start() time.Time {
	return i.getTimeField("DTSTART")
}

func (i ICal) End() time.Time {
	return i.getTimeField("DTEND")
}

func (i *ICal) UnmarshalXML(d *xml.Decoder, start xml.StartElement) error {
	s := ""
	err := d.DecodeElement(&s, &start)
	if err != nil {
		return err
	}
	ical, err := ParseICal(strings.NewReader(s))
	if err != nil {
		return err
	}

	*i = *ical
	return nil
}

func (i ICal) String() string {
	sb := &strings.Builder{}
	for n, line := range i.data {
		if n > 0 {
			sb.WriteRune('\n')
		}
		sb.WriteString(line)
	}
	return sb.String()
}

func ParseICal(r io.Reader) (*ICal, error) {
	var data []string
	scanner := bufio.NewScanner(r)
	eventStart := -1
	eventEnd := -1
	i := 0
	for scanner.Scan() {
		line := scanner.Text()
		data = append(data, line)

		if eventStart == -1 && line == "BEGIN:VEVENT" {
			eventStart = i
		}
		if eventEnd == -1 && line == "END:VEVENT" {
			eventEnd = i
		}

		i++
	}
	err := scanner.Err()
	if err != nil {
		return nil, err
	}

	// If we didn't find a start or an end, abort.
	if eventStart < 0 || eventEnd <= eventStart {
		return nil, errors.New("calendar item without event")
	}

	return &ICal{
		data:      data,
		eventData: data[eventStart+1 : eventEnd],
	}, err
}

func CreateICal(uid, summary string, start, end time.Time) *ICal {
	model := struct {
		UID     string
		Summary string
		Start   string
		End     string
		Now     string
	}{
		UID:     uid,
		Summary: summary,
		Start:   start.UTC().Format(ICAL_TIME),
		End:     end.UTC().Format(ICAL_TIME),
		Now:     time.Now().UTC().Format(ICAL_TIME),
	}

	b := &bytes.Buffer{}
	err := icalTemplate.Execute(b, model)
	if err != nil {
		// Internal stuff, so this really should not fail.
		panic(err)
	}
	ical, err := ParseICal(b)
	if err != nil {
		// Internal stuff, so this really should not fail.
		panic(err)
	}
	return ical
}

func NewCalDAV(settings ServerSettings) *CalDAV {
	return &CalDAV{
		httpClient:     http.DefaultClient,
		ServerSettings: settings,
	}
}

type CalDAVItem struct {
	HRef string
	ICal
}

func (c *CalDAV) GetEvents() ([]CalDAVItem, error) {
	req, err := http.NewRequest("PROPFIND", c.URL, strings.NewReader(`<propfind xmlns='DAV:'><prop><calendar-data xmlns='urn:ietf:params:xml:ns:caldav'/></prop></propfind>`))
	if err != nil {
		return nil, err
	}
	req.Header.Set("Content-Type", "text/xml")
	req.SetBasicAuth(c.Username, c.Password)

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return nil, err
	}

	defer resp.Body.Close()
	ms := &MultiStatus{}
	err = xml.NewDecoder(resp.Body).Decode(ms)
	if err != nil {
		return nil, err
	}

	result := make([]CalDAVItem, 0, len(ms.Response))
	for _, p := range ms.Response {
		item := CalDAVItem{p.HRef, p.PropStat.Prop.CalendarData}
		result = append(result, item)
	}

	return result, nil
}

func (c CalDAV) UploadItem(item CalDAVItem) error {
	itemURL, _ := url.Parse(item.HRef)
	base, _ := url.Parse(c.URL)

	// Build the absolute URL for the item
	itemURL = base.ResolveReference(itemURL)

	req, err := http.NewRequest("PUT", itemURL.String(), strings.NewReader(item.ICal.String()))
	if err != nil {
		return err
	}
	req.SetBasicAuth(c.Username, c.Password)
	resp, err := c.httpClient.Do(req)
	if err != nil {
		return err
	}

	if resp.StatusCode >= 300 {
		return fmt.Errorf("Unexpected result while uploading %s: %d", itemURL.String(), resp.StatusCode)
	}

	return nil
}

func (c CalDAV) DeleteItem(item CalDAVItem) error {
	itemURL, _ := url.Parse(item.HRef)
	base, _ := url.Parse(c.URL)

	// Build the absolute URL for the item
	itemURL = base.ResolveReference(itemURL)

	req, err := http.NewRequest("DELETE", itemURL.String(), nil)
	if err != nil {
		return err
	}
	req.SetBasicAuth(c.Username, c.Password)
	resp, err := c.httpClient.Do(req)
	if err != nil {
		return err
	}

	if resp.StatusCode >= 300 {
		return fmt.Errorf("Unexpected result while deleting %s: %d", itemURL.String(), resp.StatusCode)
	}

	return nil
}

type MultiStatus struct {
	Response []PropFindResponse `xml:"response"`
}

type PropFindResponse struct {
	HRef     string `xml:"href"`
	PropStat struct {
		Prop struct {
			CalendarData ICal `xml:"calendar-data"`
		} `xml:"prop"`
	} `xml:"propstat"`
}

var icalTemplate = template.Must(template.New("icalTemplate").Parse(`BEGIN:VCALENDAR
PRODID:-//aksdb//calanonsync//EN
VERSION:2.0
BEGIN:VEVENT
UID:{{ .UID }}
SUMMARY:{{ .Summary }}
CLASS:PUBLIC
DTSTART:{{ .Start }}Z
DTEND:{{ .End }}Z
CREATED:{{ .Now }}Z
DTSTAMP:{{ .Now }}Z
LAST-MODIFIED:{{ .Now }}Z
END:VEVENT
END:VCALENDAR`))
